@@ -1,5 +1,6 @@
 /*
  *  Copyright (C) 2010 Felix Geyer <debfx@fobos.de>
+ *  Copyright (C) 2017 KeePassXC Team <team@keepassxc.org>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -75,8 +76,11 @@ EditEntryWidget::EditEntryWidget(QWidget* parent)
     setupProperties();
     setupHistory();
 
-    connect(this, SIGNAL(accepted()), SLOT(saveEntry()));
+    connect(this, SIGNAL(accepted()), SLOT(acceptEntry()));
     connect(this, SIGNAL(rejected()), SLOT(cancel()));
+    connect(this, SIGNAL(apply()), SLOT(saveEntry()));
+    connect(m_iconsWidget, SIGNAL(messageEditEntry(QString, MessageWidget::MessageType)), SLOT(showMessage(QString, MessageWidget::MessageType)));
+    connect(m_iconsWidget, SIGNAL(messageEditEntryDismiss()), SLOT(hideMessage()));
     
     m_mainUi->passwordGenerator->layout()->setContentsMargins(0, 0, 0, 0);
 }
@@ -88,7 +92,7 @@ EditEntryWidget::~EditEntryWidget()
 void EditEntryWidget::setupMain()
 {
     m_mainUi->setupUi(m_mainWidget);
-    add(tr("Entry"), m_mainWidget);
+    addPage(tr("Entry"), FilePath::instance()->icon("actions", "document-edit"), m_mainWidget);
 
     m_mainUi->togglePasswordButton->setIcon(filePath()->onOffIcon("actions", "password-show"));
     m_mainUi->togglePasswordGeneratorButton->setIcon(filePath()->icon("actions", "password-generator", false));
@@ -113,7 +117,7 @@ void EditEntryWidget::setupMain()
 void EditEntryWidget::setupAdvanced()
 {
     m_advancedUi->setupUi(m_advancedWidget);
-    add(tr("Advanced"), m_advancedWidget);
+    addPage(tr("Advanced"), FilePath::instance()->icon("categories", "preferences-other"), m_advancedWidget);
 
     m_attachmentsModel->setEntryAttachments(m_entryAttachments);
     m_advancedUi->attachmentsView->setModel(m_attachmentsModel);
@@ -130,6 +134,8 @@ void EditEntryWidget::setupAdvanced()
     connect(m_advancedUi->addAttributeButton, SIGNAL(clicked()), SLOT(insertAttribute()));
     connect(m_advancedUi->editAttributeButton, SIGNAL(clicked()), SLOT(editCurrentAttribute()));
     connect(m_advancedUi->removeAttributeButton, SIGNAL(clicked()), SLOT(removeCurrentAttribute()));
+    connect(m_advancedUi->protectAttributeButton, SIGNAL(toggled(bool)), SLOT(protectCurrentAttribute(bool)));
+    connect(m_advancedUi->revealAttributeButton, SIGNAL(clicked(bool)), SLOT(revealCurrentAttribute()));
     connect(m_advancedUi->attributesView->selectionModel(),
             SIGNAL(currentChanged(QModelIndex,QModelIndex)),
             SLOT(updateCurrentAttribute()));
@@ -137,13 +143,13 @@ void EditEntryWidget::setupAdvanced()
 
 void EditEntryWidget::setupIcon()
 {
-    add(tr("Icon"), m_iconsWidget);
+    addPage(tr("Icon"), FilePath::instance()->icon("apps", "preferences-desktop-icons"), m_iconsWidget);
 }
 
 void EditEntryWidget::setupAutoType()
 {
     m_autoTypeUi->setupUi(m_autoTypeWidget);
-    add(tr("Auto-Type"), m_autoTypeWidget);
+    addPage(tr("Auto-Type"), FilePath::instance()->icon("actions", "key-enter"), m_autoTypeWidget);
 
     m_autoTypeDefaultSequenceGroup->addButton(m_autoTypeUi->inheritSequenceButton);
     m_autoTypeDefaultSequenceGroup->addButton(m_autoTypeUi->customSequenceButton);
@@ -175,13 +181,13 @@ void EditEntryWidget::setupAutoType()
 
 void EditEntryWidget::setupProperties()
 {
-    add(tr("Properties"), m_editWidgetProperties);
+    addPage(tr("Properties"), FilePath::instance()->icon("actions", "document-properties"), m_editWidgetProperties);
 }
 
 void EditEntryWidget::setupHistory()
 {
     m_historyUi->setupUi(m_historyWidget);
-    add(tr("History"), m_historyWidget);
+    addPage(tr("History"), FilePath::instance()->icon("actions", "view-history"), m_historyWidget);
 
     m_sortModel->setSourceModel(m_historyModel);
     m_sortModel->setDynamicSortFilter(true);
@@ -208,7 +214,7 @@ void EditEntryWidget::emitHistoryEntryActivated(const QModelIndex& index)
     Q_ASSERT(!m_history);
 
     Entry* entry = m_historyModel->entryFromIndex(index);
-    Q_EMIT historyEntryActivated(entry);
+    emit historyEntryActivated(entry);
 }
 
 void EditEntryWidget::histEntryActivated(const QModelIndex& index)
@@ -289,8 +295,8 @@ void EditEntryWidget::loadEntry(Entry* entry, bool create, bool history, const Q
     setForms(entry);
     setReadOnly(m_history);
 
-    setCurrentRow(0);
-    setRowHidden(m_historyWidget, m_history);
+    setCurrentPage(0);
+    setPageHidden(m_historyWidget, m_history || m_entry->historyItems().count() < 1);
 }
 
 void EditEntryWidget::setForms(const Entry* entry, bool restore)
@@ -349,6 +355,11 @@ void EditEntryWidget::setForms(const Entry* entry, bool restore)
         m_advancedUi->attributesEdit->setEnabled(false);
     }
 
+    QList<int> sizes = m_advancedUi->attributesSplitter->sizes();
+    sizes.replace(0, m_advancedUi->attributesSplitter->width() * 0.3);
+    sizes.replace(1, m_advancedUi->attributesSplitter->width() * 0.7);
+    m_advancedUi->attributesSplitter->setSizes(sizes);
+
     IconStruct iconStruct;
     iconStruct.uuid = entry->iconUuid();
     iconStruct.number = entry->iconNumber();
@@ -397,16 +408,17 @@ void EditEntryWidget::saveEntry()
 {
     if (m_history) {
         clear();
-        Q_EMIT editFinished(false);
+        hideMessage();
+        emit editFinished(false);
         return;
     }
 
     if (!passwordsEqual()) {
-        MessageBox::warning(this, tr("Error"), tr("Different passwords supplied."));
+        showMessage(tr("Different passwords supplied."), MessageWidget::Error);
         return;
     }
 
-    if (m_advancedUi->attributesView->currentIndex().isValid()) {
+    if (m_advancedUi->attributesView->currentIndex().isValid() && m_advancedUi->attributesEdit->isEnabled()) {
         QString key = m_attributesModel->keyByIndex(m_advancedUi->attributesView->currentIndex());
         m_entryAttributes->set(key, m_advancedUi->attributesEdit->toPlainText(),
                                m_entryAttributes->isProtected(key));
@@ -429,10 +441,13 @@ void EditEntryWidget::saveEntry()
     if (!m_create) {
         m_entry->endUpdate();
     }
+}
 
+void EditEntryWidget::acceptEntry()
+{
+    saveEntry();
     clear();
-
-    Q_EMIT editFinished(true);
+    emit editFinished(true);
 }
 
 void EditEntryWidget::updateEntryData(Entry* entry) const
@@ -476,7 +491,8 @@ void EditEntryWidget::cancel()
 {
     if (m_history) {
         clear();
-        Q_EMIT editFinished(false);
+        hideMessage();
+        emit editFinished(false);
         return;
     }
 
@@ -487,7 +503,7 @@ void EditEntryWidget::cancel()
 
     clear();
 
-    Q_EMIT editFinished(false);
+    emit editFinished(false);
 }
 
 void EditEntryWidget::clear()
@@ -499,6 +515,7 @@ void EditEntryWidget::clear()
     m_autoTypeAssoc->clear();
     m_historyModel->clear();
     m_iconsWidget->reset();
+    hideMessage();
 }
 
 bool EditEntryWidget::hasBeenModified() const
@@ -573,46 +590,96 @@ void EditEntryWidget::removeCurrentAttribute()
     QModelIndex index = m_advancedUi->attributesView->currentIndex();
 
     if (index.isValid()) {
-        m_entryAttributes->remove(m_attributesModel->keyByIndex(index));
+        if (MessageBox::question(this, tr("Confirm Remove"), tr("Are you sure you want to remove this attribute?"),
+                                 QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes) {
+            m_entryAttributes->remove(m_attributesModel->keyByIndex(index));
+        }
     }
 }
 
 void EditEntryWidget::updateCurrentAttribute()
 {
     QModelIndex newIndex = m_advancedUi->attributesView->currentIndex();
+    QString newKey = m_attributesModel->keyByIndex(newIndex);
 
-    if (m_history) {
-        if (newIndex.isValid()) {
-            QString key = m_attributesModel->keyByIndex(newIndex);
-            m_advancedUi->attributesEdit->setPlainText(m_entryAttributes->value(key));
-            m_advancedUi->attributesEdit->setEnabled(true);
+    if (!m_history && m_currentAttribute != newIndex) {
+        // Save changes to the currently selected attribute if editing is enabled
+        if (m_currentAttribute.isValid() && m_advancedUi->attributesEdit->isEnabled()) {
+            QString currKey = m_attributesModel->keyByIndex(m_currentAttribute);
+            m_entryAttributes->set(currKey, m_advancedUi->attributesEdit->toPlainText(),
+                                   m_entryAttributes->isProtected(currKey));
+        }        
+    }
+
+    displayAttribute(newIndex, m_entryAttributes->isProtected(newKey));
+
+    m_currentAttribute = newIndex;
+}
+
+void EditEntryWidget::displayAttribute(QModelIndex index, bool showProtected)
+{
+    // Block signals to prevent extra calls
+    m_advancedUi->protectAttributeButton->blockSignals(true);
+
+    if (index.isValid()) {
+        QString key = m_attributesModel->keyByIndex(index);
+        if (showProtected) {
+            m_advancedUi->attributesEdit->setPlainText(tr("[PROTECTED] Press reveal to view or edit"));
+            m_advancedUi->attributesEdit->setEnabled(false);
+            m_advancedUi->revealAttributeButton->setEnabled(true);
+            m_advancedUi->protectAttributeButton->setChecked(true);
         }
         else {
-            m_advancedUi->attributesEdit->setPlainText("");
-            m_advancedUi->attributesEdit->setEnabled(false);
+            m_advancedUi->attributesEdit->setPlainText(m_entryAttributes->value(key));
+            m_advancedUi->attributesEdit->setEnabled(true);
+            m_advancedUi->revealAttributeButton->setEnabled(false);
+            m_advancedUi->protectAttributeButton->setChecked(false);
         }
+
+        // Don't allow editing in history view
+        m_advancedUi->protectAttributeButton->setEnabled(!m_history);
+        m_advancedUi->editAttributeButton->setEnabled(!m_history);
+        m_advancedUi->removeAttributeButton->setEnabled(!m_history);
     }
     else {
-        if (m_currentAttribute != newIndex) {
-            if (m_currentAttribute.isValid()) {
-                QString key = m_attributesModel->keyByIndex(m_currentAttribute);
-                m_entryAttributes->set(key, m_advancedUi->attributesEdit->toPlainText(),
-                                       m_entryAttributes->isProtected(key));
-            }
+        m_advancedUi->attributesEdit->setPlainText("");
+        m_advancedUi->attributesEdit->setEnabled(false);
+        m_advancedUi->revealAttributeButton->setEnabled(false);
+        m_advancedUi->protectAttributeButton->setChecked(false);
+        m_advancedUi->protectAttributeButton->setEnabled(false);
+        m_advancedUi->editAttributeButton->setEnabled(false);
+        m_advancedUi->removeAttributeButton->setEnabled(false);
+    }
 
-            if (newIndex.isValid()) {
-                QString key = m_attributesModel->keyByIndex(newIndex);
-                m_advancedUi->attributesEdit->setPlainText(m_entryAttributes->value(key));
-                m_advancedUi->attributesEdit->setEnabled(true);
-            }
-            else {
-                m_advancedUi->attributesEdit->setPlainText("");
-                m_advancedUi->attributesEdit->setEnabled(false);
-            }
+    m_advancedUi->protectAttributeButton->blockSignals(false);
+}
 
-            m_advancedUi->editAttributeButton->setEnabled(newIndex.isValid());
-            m_advancedUi->removeAttributeButton->setEnabled(newIndex.isValid());
-            m_currentAttribute = newIndex;
+void EditEntryWidget::protectCurrentAttribute(bool state)
+{
+    QModelIndex index = m_advancedUi->attributesView->currentIndex();
+    if (!m_history && index.isValid()) {
+        QString key = m_attributesModel->keyByIndex(index);
+        if (state) {
+            // Save the current text and protect the attribute
+            m_entryAttributes->set(key, m_advancedUi->attributesEdit->toPlainText(), true);
+        } else {
+            // Unprotect the current attribute value (don't save text as it is obscured)
+            m_entryAttributes->set(key, m_entryAttributes->value(key), false);
+        }
+
+        // Display the attribute
+        displayAttribute(index, state);
+    }
+}
+
+void EditEntryWidget::revealCurrentAttribute()
+{
+    if (! m_advancedUi->attributesEdit->isEnabled()) {
+        QModelIndex index = m_advancedUi->attributesView->currentIndex();
+        if (index.isValid()) {
+            QString key = m_attributesModel->keyByIndex(index);
+            m_advancedUi->attributesEdit->setPlainText(m_entryAttributes->value(key));
+            m_advancedUi->attributesEdit->setEnabled(true);
         }
     }
 }
@@ -632,15 +699,13 @@ void EditEntryWidget::insertAttachment()
 
     QFile file(filename);
     if (!file.open(QIODevice::ReadOnly)) {
-        MessageBox::warning(this, tr("Error"),
-                tr("Unable to open file").append(":\n").append(file.errorString()));
+        showMessage(tr("Unable to open file").append(":\n").append(file.errorString()), MessageWidget::Error);
         return;
     }
 
     QByteArray data;
     if (!Tools::readAllFromDevice(&file, data)) {
-        MessageBox::warning(this, tr("Error"),
-                tr("Unable to open file").append(":\n").append(file.errorString()));
+        showMessage(tr("Unable to open file").append(":\n").append(file.errorString()), MessageWidget::Error);
         return;
     }
 
@@ -667,13 +732,11 @@ void EditEntryWidget::saveCurrentAttachment()
 
         QFile file(savePath);
         if (!file.open(QIODevice::WriteOnly)) {
-            MessageBox::warning(this, tr("Error"),
-                    tr("Unable to save the attachment:\n").append(file.errorString()));
+            showMessage(tr("Unable to save the attachment:\n").append(file.errorString()), MessageWidget::Error);
             return;
         }
         if (file.write(attachmentData) != attachmentData.size()) {
-            MessageBox::warning(this, tr("Error"),
-                    tr("Unable to save the attachment:\n").append(file.errorString()));
+            showMessage(tr("Unable to save the attachment:\n").append(file.errorString()), MessageWidget::Error);
             return;
         }
     }
@@ -694,20 +757,17 @@ void EditEntryWidget::openAttachment(const QModelIndex& index)
     QTemporaryFile* file = new QTemporaryFile(tmpFileTemplate, this);
 
     if (!file->open()) {
-        MessageBox::warning(this, tr("Error"),
-                tr("Unable to save the attachment:\n").append(file->errorString()));
+        showMessage(tr("Unable to save the attachment:\n").append(file->errorString()), MessageWidget::Error);
         return;
     }
 
     if (file->write(attachmentData) != attachmentData.size()) {
-        MessageBox::warning(this, tr("Error"),
-                tr("Unable to save the attachment:\n").append(file->errorString()));
+        showMessage(tr("Unable to save the attachment:\n").append(file->errorString()), MessageWidget::Error);
         return;
     }
 
     if (!file->flush()) {
-        MessageBox::warning(this, tr("Error"),
-                tr("Unable to save the attachment:\n").append(file->errorString()));
+        showMessage(tr("Unable to save the attachment:\n").append(file->errorString()), MessageWidget::Error);
         return;
     }
 
@@ -732,8 +792,13 @@ void EditEntryWidget::removeCurrentAttachment()
         return;
     }
 
-    QString key = m_attachmentsModel->keyByIndex(index);
-    m_entryAttachments->remove(key);
+    QMessageBox::StandardButton ans = MessageBox::question(this, tr("Confirm Remove"),
+                                                           tr("Are you sure you want to remove this attachment?"),
+                                                           QMessageBox::Yes | QMessageBox::No);
+    if (ans == QMessageBox::Yes) {
+        QString key = m_attachmentsModel->keyByIndex(index);
+        m_entryAttachments->remove(key);
+    }
 }
 
 void EditEntryWidget::updateAutoTypeEnabled()
